"""A super-fast user agent string parser"""


__version__ = '0.1.3'


def parse_ua(s, always_include_os=False):
    """ Parse the given user-agent string and return a string of the
    form "client - os". Use x.partition(' - ') if needed. The
    client is "Browser" if unknown. The os is "Other" if unknown and
    may contain a string "Mobile", "Tablet" or "Desktop. When the ua
    does not represent a browser (but e.g. a bot or wget) the os is
    omitted (unless always_include_os is True).
    """
    s = str(s)
    ls = s.lower()
    is_browser = s.startswith("Mozilla/5")
    client = opsys = device_type = ""

    if not is_browser:
        # This is a bot or something like wget (or a weird/old browser)
        client = "Other"
        if "bot" in ls or "crawl" in ls or "spider" in ls or "scrap" in ls:
            client = "Bot"
        elif "indexer" in ls or "pinger" in ls or "monitor" in s:
            client = "Bot"
        elif "facebook" in ls or "google" in ls or "Argus" in s:
            client = "Bot"

    elif "bot" in ls or "crawl" in ls or "scraper" in ls or "spider" in ls:
        # This is a bot pretending to be a browser
        is_browser = False
        client = "Bot"

    else:
        # This is a browser, try to detect which one!
        client = "Browser"

        if "Firefox" in s:
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/User-Agent/Firefox
            client = "Firefox"
            if "Gecko/20100101" in s:
                # On Desktop, geckotrail is the fixed string "20100101"
                device_type = "Desktop"
            else:
                # Mobile!
                device_type = "Mobile"
                if "Mobi" in s:
                    device_type = "Mobile"  # but now we're sure :)
                elif "Tablet" in s:
                    device_type = "Tablet"
            # Some clients include Firefox for compat
            # Some are based on Firefox. Perhaps we should just call them FF
            if "Trident" in s:
                client = "IE"
            # elif "PaleMoon" in s:
            #     client = "Pale Moon"
            # elif "SailfishBrowser" in s:
            #     client = "Sailfish Browser"
            # elif "IceWeasel" in s:
            #     client = "IceWeasel"
            # elif "Waterfox" in s:
            #     client = "Waterfox"
            # elif "Basilisk" in s:
            #     client = "Basilisk"
        elif "Edg" in s and any(x in s for x in ("Edge", "Edg/", "EdgiOS/", "EdgA/")):
            client = "Edge"
        elif "Chrom" in s:
            # https://developer.chrome.com/multidevice/user-agent
            # Note that "Safari" is also present, so check for Chrome first
            client = "Chrome"
            if "HeadlessChrome" in s:
                client = "HeadlessChrome"  # Can be Chrome or Chromium based
            elif "Chromium" in s:
                client = "Chromium"
            elif "Brave" in s:
                client = "Brave"
            elif "OPR" in s:
                client = "Opera"  # put here because it includes Chrome for compat
            elif "Silk" in s:
                client = "Silk"
        elif "Safari" in s:
            client = "Safari"
            if "FxiOS" in s:
                client = "Firefox"
            elif "CriOS" in s:
                client = "Chrome"
            elif "Coast" in s or "OPiOS" in s:
                client = "Opera"
            elif "Silk" in s:
                client = "Silk"
        elif "IE" in s or "Trident" in s:
            client = "IE"

        if not device_type and "mobi" in ls:
            device_type = "Mobile"

    # Try to detect the OS if this is a browser
    if is_browser or always_include_os:
        opsys = "Other"
        if "Windows" in s:
            opsys = "Windows"
            if "X11" in s:
                opsys = "Chrome OS"  # WTF
            if not device_type:
                if "Zune" in s or "Phone" in s:
                    device_type = "Mobile"
        elif "Android" in s:
            opsys = "Android"
            if "Bot" in s:
                opsys = "Other"
            elif "iOS" in s:
                opsys = "iOS"
        elif "watchOS" in s:
            opsys = "WatchOS"
        elif "kaios" in ls:
            opsys = "KaiOS"
        elif "Silk-Accelerated" in s:
            opsys = "Android"
        elif "BSD" in s:
            opsys = "BSD"  # includes FreeBSD, NetBSD and OpenBSD
        elif "Kindle" in s:  # check before Linux
            opsys = "Kindle"
        elif "linux" in ls:
            opsys = "Linux"
            if "Sailfish" in s:
                opsys = "Sailfish"  # Adroidish os
            elif "Mobile" in s or "UCW" in s:
                opsys = "Android"
            elif "Philips" in s or "Sony" in s or "LG" in s:
                opsys = "Other"
            elif "TV" in s or "wOSBrowser" in s or "WeTab" in s or "CrKey" in s:
                opsys = "Other"
            elif "Maemo" in s:
                opsys = "Other"
        elif "SunOS" in s:
            opsys = "Solaris"
        elif "iOS" in s:
            opsys = "iOS"
        elif "iP" in s and ("iPad" in s or "iPod" in s or "iPhone" in s):
            opsys = "iOS"
        elif "Mac" in s:
            opsys = "Mac OS X"
            if "Mac_PowerPC" in s:
                opsys = "Mac OS"
            elif "Apple TV" in s:
                opsys = "Other"
        elif "Darwin" in s:
            if "86" in s:  # x86 or i86
                opsys = "Mac OS X"
            else:
                opsys = "iOS"
        elif "Nokia" in s:
            opsys = "Nokia"
        elif "BB10" in s or "BlackBerry" in s:
            opsys = "BlackBerry OS"
        elif "PlayBook" in s:
            opsys = "BlackBerry Tablet OS"

    # Return as str
    if device_type:
        opsys = opsys + " " + device_type
    if opsys:
        client = client + " - " + opsys
    return client



"""
Tests for fastuaparser.
"""

import os
import json
import time

from fastuaparser import parse_ua

ua_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ua_data.json")
if not os.path.isfile(ua_file):
    raise RuntimeError("ua_data.json does not exist, maybe run generator script first?")


# Some uas queried with my machine
uas = {
    "Firefox": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:66.0) Gecko/20100101 Firefox/66.0",
    "Chrome": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36",
    "Edge": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.140 Safari/537.36 Edge/17.17134",
    "IE": "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko",
}


# Clients based on Firefox that we chose to consider "Firefox"
firefox_like = "Pale Moon", "IceWeasel", "Waterfox", "Sailfish Browser", "Basilisk"

ie_like = "QQ Browser", "AOL"

# Clients that are ok if we call them "Browser"
browser_like = ("Pale Moon",)

clients_of_interest = (
    "Firefox",
    "Chrome",
    "Chromium",
    "Android browser",
    "Edge",
    "IE",
    "Safari",
    "Opera",
    "Brave",
    "Silk",
)

# Some ua's dont start with Mozilla/5, these are either ancient browsers
# or (old?) Opera browsers on certain devices.
silly_ua_starts = (
    "Mozilla/4",
    "Opera",
    "(Opera",
    "HTC",
    "LG",
    "Samsung",
    "SAMSUNG",
    "Safari",
)

# Some bots that we sometimes mark as Other, or Bot, or Firefox, or ...
crap_bots = (
    "Google-HTTP-Java-Client",
    "Daumoa",
    "Daumoa-feedfetcher",
    "Facebook",
    "Box Sync",
    "Google",
    "Web-sniffer",
    "WebPageTest.org bot",
)

# If any of these words are present in the ref client name, we consider it a bot
bot_hints = "bot", "crawl", "spider", "preview", "scraper"

# Names of bots that are not detected by the above. We just call all these "Bot"
bot_like = """
008, 50.nu, A6-Indexer, Aboundex, Accoona-AI-Agent, Accoona-Biz-Agent,
Altresium, Android, Ant-Nutch, AppEngine-Google, Argus, Ask Jeeves, Axtaris,
BaiduMobaider, BoardReader, BoardReader Blog Indexer, Checklinks, DNSGroup,
EDI, Flamingo_SearchEngine, GmailImageProxy, Goodzer, GoogleImageProxy, Grub,
HiddenMarket, HooWWWer, INGRID, IconSurf, IlTrovatore, Infohelfer, InfuzApp,
InternetArchive, Isara, Kraken, Kurzor, LEIA, LOOQ, LinkAider, Llaut,
Mediapartners-Google, NewsGator, Nutch, Nutch; http:, NutchCVS, NutchOSUOSL,
NutchOrg, ObjectsSearch, Orbiter, Other, PagesInventory, PathDefender,
PaxleFramework, Peew, PostPost, RedCarpet, Scrapy, ShopSalad, Simpy,
Slack-ImgProxy, Steeler, Tailsweep, Thumbshots.ru, VSE, Vagabondo, WIRE,
WebCrunch, WebIndexer, WebZIP, Wotbox, Y!J-BRI, Y!J-BRW, YahooSeeker, Yeti,
Yowedo, Zao, ZyBorg, aws-cli, boitho.com-dc, clumboot, envolk, gonzo1,
grub-client, heritrix, holmes, ichiro, ichiro/mobile, mozDex,
semanticdiscovery, sproose, voyager, wminer, www.almaden.ibm.com,
Charlotte
"""
bot_like = [x.strip() for x in bot_like.strip().replace("\n", " ").split(", ")]

linux_like = "Debian", "Ubuntu", "Mandriva", "Gentoo", "Red Hat"
bsd_like = "OpenBSD", "FreeBSD", "NetBSD"

nokia_like = (
    "MeeGo",
    "Nokia Series 40",
    "Other",
    "Symbian OS",
    "Symbian^3",
    "Symbian^3 Anna",
    "Symbian^3 Belle",
    "Nokia Series 30 Plus",
)


oses_not_of_interest = (
    "Bada",
    "Brew",
    "BREW",
    "Brew MP",
    "Symbian OS",
    "VRE",
    "webOS",
    "WebTV",
    "Samsung",
    "Philips",
    "LG",
    "Sony",
    "webOS",
    "WeTab",
    "Panasonic",
    "FireHbbTV",
    "Maemo",
    "GoogleTV",
    "ATV OS X",
    "tvOS",
    "Roku",
    "Web0S",
    "Tizen",
    "Chromecast",
    "Firefox OS",
)


def test_parse_ua_basics():

    # Does not raise
    assert parse_ua(None) == "Other"
    assert parse_ua("") == "Other"
    assert parse_ua(4) == "Other"

    # Check for common ua's
    for browser, ua in uas.items():
        assert parse_ua(ua) in (browser + " - Windows", browser + " - Windows Desktop")

    # Check that this first bit is important
    for browser, ua in uas.items():
        assert parse_ua(ua[1:]) == "Other"
        assert parse_ua("x" + ua) == "Other"

    # Unknown browsers are called "Browser
    assert parse_ua("Mozilla/5.0 (Linux) X") == "Browser - Linux"
    assert parse_ua("Mozilla/5.0 (Y) X") == "Browser - Other"

    # Unknown os's are called "Other"
    for browser, ua in uas.items():
        assert parse_ua(ua.replace("Win", "Loose")) in (
            browser + " - Other",
            browser + " - Other Desktop",
        )

    # We detect bots
    for browser, ua in uas.items():
        assert parse_ua(ua + "bot") == "Bot"

    # ... in several ways
    for ua in ("xxspider", "crawlee", "scrapy", "indexeroo"):
        assert parse_ua(ua) == "Bot"


def speedometer():
    # Read test data
    with open(ua_file, "rb") as f:
        cases = json.loads(f.read().decode())

    # Read as fast as we can
    t0 = time.perf_counter()
    for case in cases:
        parse_ua(case["ua"])
    t1 = time.perf_counter()
    print(len(cases), "ua strings in", t1 - t0, "s")


def test_parseua_client():

    # Read test data
    with open(ua_file, "rb") as f:
        cases = json.loads(f.read().decode())
    cases = [case for case in cases if case.get("client", None)]

    # Some more preparations
    clientmap = {}
    for case in cases:
        ref_client = case["client"]
        for x in clients_of_interest:
            if x in ref_client:
                clientmap[ref_client] = x
                break
    for x in firefox_like:
        clientmap[x] = "Firefox"
    for ign in (
        "Mail.ru Chromium Browser",
        "Firefox (Shiretoko)",
        "Mobile Safari UI/WKWebView",
        "HeadlessChrome",
        "Opera Neon",
    ):
        clientmap.pop(ign)

    # Test client
    for case in cases:
        s = case["ua"]
        ref_client = case["client"]
        ref_alias = clientmap.get(ref_client, None)
        client, _, os = parse_ua(s).partition(" - ")
        if client != ref_client:
            if ref_client in crap_bots:
                pass
            elif client == "Bot" and any(x in ref_client.lower() for x in bot_hints):
                pass  # The ref client really looks like a bot
            elif client == "Bot" and ref_client in bot_like:
                pass  # We've marked this ref client as a bot
            elif ref_alias is None and client in ("Other", "Browser"):
                pass  # A browser that we're not interested in
            elif client == ref_alias:
                pass  # alias, we use simpler names (are less specific)
            elif client == "Other" and s.startswith(silly_ua_starts):
                pass  # Some real old browser, meh
            elif client == "Browser" and ref_client in browser_like:
                pass
            elif ref_client == "Android" and client in ("Safari", "Chrome"):
                pass  # builtin android browser is based on either
            elif client == "Chrome" and ref_alias is None:
                pass  # A Chrome-based browser, these include Electron apps
            elif client == "Safari" and ref_alias is None:
                pass  # A Safari based browser
            elif client == "IE" and ref_client in ie_like:
                pass
            else:
                assert False, (client, ref_client)


def test_parse_ua_os():

    # Read test data
    with open(ua_file, "rb") as f:
        cases = json.loads(f.read().decode())
    cases = [case for case in cases if case.get("os", None)]

    # Test os
    for case in cases:
        s = case["ua"]
        ref_os = case["os"]
        client, _, os = parse_ua(s, True).partition(" - ")
        os = os.replace(" Mobile", "").replace(" Tablet", "").replace(" Desktop", "")
        ref_os = (
            ref_os.replace(" Mobile", "").replace(" Tablet", "").replace(" Phone", "")
        )
        if os != ref_os:
            if os == "Linux" and ref_os in linux_like:
                pass
            elif os == "BSD" and ref_os in bsd_like:
                pass
            elif os == "Nokia" and ref_os in nokia_like:
                pass
            elif os == "Other" and (s.startswith("Wget") or s.startswith("curl")):
                pass  # These use lowercase for Linux/BSD, meh
            elif os == "Other" and s.startswith("SalesforceMobileSDK"):
                pass  # Special case
            elif os == "Other" and ref_os in oses_not_of_interest:
                pass
            else:
                assert False, (os, ref_os)
        else:
            pass


if __name__ == "__main__":
    test_parse_ua_basics()
    test_parseua_client()
    test_parse_ua_os()

    speedometer()




"""
Script to generate test data for parsing ua strings. Most data is from the
ua-parser project (https://github.com/ua-parser/uap-core), but we also
include a few ua's of our own.
"""

import os
import json
from urllib.request import urlopen

import yaml


dir = os.path.dirname(os.path.abspath(__file__))
url = "https://raw.githubusercontent.com/ua-parser/uap-core/master/tests"


def generate_data():

    data = {}  # ua -> {}

    # Read os data from ua-parser project
    with urlopen(url + "/test_os.yaml") as f:
        os_data = yaml.load(f)
    for case in os_data["test_cases"]:
        d = data.setdefault(case["user_agent_string"], {})
        d["os"] = case["family"]

    # Read client data from ua-parser project
    with urlopen(url + "/test_ua.yaml") as f:
        client_data = yaml.load(f)
    for case in client_data["test_cases"]:
        ua = case["user_agent_string"]
        if ua.startswith("User agent:"):  # error in the data
            ua = ua.split(":", 1)[-1].lstrip()
        d = data.setdefault(ua, {})
        d["client"] = case["family"]

    # Read data from ourselves
    for case in more_cases:
        d = data.setdefault(case["ua"], {})
        d["client"] = case["client"]
        d["os"] = case["os"]

    # Turn into a list
    test_cases = []
    for ua, d in data.items():
        d["ua"] = ua
        test_cases.append(d)

    # Write
    with open(os.path.join(dir, "ua_data.json"), "wb") as f:
        f.write(json.dumps(test_cases, indent=2).encode())


more_cases = [
    {
        "ua": "Mozilla/5.0 (Windows  10.0; Win64; x64; rv:66.0) Gecko/20100101 Firefox/66.0",
        "client": "Firefox",
        "os": "Windows",
    },
    {
        "ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:66.0) Gecko/20100101 Firefox/66.0",
        "client": "Firefox",
        "os": "Windows",
    },
    {
        "ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Safari/537.36",
        "client": "Chrome",
        "os": "Windows",
    },
    {
        "ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.140 Safari/537.36 Edge/17.17134",
        "client": "Edge",
        "os": "Windows",
    },
    {
        "ua": "Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko",
        "client": "IE",
        "os": "Windows",
    },
]


if __name__ == "__main__":
    generate_data()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='fastuaparser',
      version='0.1.3',
      description='A super-fast user agent string parser',
      author='Almar Klein',
      author_email='almar.klein@gmail.com',
      url='https://github.com/almarklein/fastuaparser',
      py_modules=['fastuaparser'],
     )
